"""
View module for pyegeria, containing output formatters and mermaid utilities.
"""
from pyegeria.view.mermaid_utilities import (
    construct_mermaid_web,
    construct_mermaid_jup,
    load_mermaid,
    render_mermaid,
    save_mermaid_html,
    save_mermaid_graph,
)
from pyegeria.view.output_formatter import (
    generate_output,
    resolve_output_formats,
    populate_common_columns,

)
from pyegeria.view.format_set_executor import (
    exec_report_spec
)
__all__ = [
    "construct_mermaid_web",
    "construct_mermaid_jup",
    "load_mermaid",
    "render_mermaid",
    "save_mermaid_html",
    "save_mermaid_graph",
    "generate_output",
    "resolve_output_formats",
    "populate_common_columns",
    "exec_report_spec"
]
